import sqlvalidator
from powerml_app.utils.utils import run_and_postprocess


def validate_sql_select(query):
    '''
    Uses library to check whether text is valid sql select statement

    query: Input text to check.
    '''
    sql_query = sqlvalidator.parse(query)
    return sql_query.is_valid()


def fuzzy_is_sql(query):
    '''
    Uses LLM to check whether text contains a valid SQL program.

    text: Input query to check.
    '''

    examples = [
        "SELECT * from customers;\n\nFrom the above text, determine if it is a valid SQL program.  Answer Yes or No.\nYes. This query is valid.",
        "SELECT CAT(month as timestamp) as month from customers;\n\nFrom the above text, determine if it is a valid SQL program.  Answer Yes or No.\nNo. CAT should be CAST."
    ]

    prompt_append = f'\n\nFrom the above text, determine if it is a valid SQL program.  Answer Yes or No, and explain why or why not.'
    prompt = "\n\n".join(examples) + query + prompt_append

    is_sql = run_and_postprocess(prompt)

    # Prepend is to avoid non-professionalism in output
    # TODO: Create and add to shared prepended prompts file

    # prompt_append = f'\n\nFrom the above text, determine if it is a valid SQL program, or if there are any syntax errors.  Answer Yes or No.'
    # prompt = query + prompt_append

    # is_sql = run_and_postprocess(prompt)

    return is_sql


if __name__ == "__main__":
    texts = [
        'I like carrots and llamas.',
        '''CX: I'd like to order.
AI: Sorry we are out of that.
CX: Sorry we are really really out of that.
Order: 1x Vanilla Syrup''',
        '''
CREATE TABLE STATION
(ID INTEGER PRIMARY KEY,
CITY CHAR(20),
STATE CHAR(2),
LAT_N REAL,
LONG_W REAL);''',
        '''
SELECT * from table;''', '''
SELECT * from table'''
    ]

    for text in texts:
        bool_check = validate_sql_select(text)
        print(bool_check)



from powerml_app.utils.utils import run_and_postprocess


def fuzzy_contains(text, check_contains, return_reason=False):
    '''
    Uses LLM to check whether text contains a concept.

    text: Input text to check against.
    check_contains: Word or phrase to check if it's contained in `text`. Case matters.
    return_reason: If the check is contained, then return the word(s) from the text that support check.
    '''

    # Prepend is to avoid non-professionalism in output
    # TODO: Create and add to shared prepended prompts file
    prompt_prepend = f'The following material is from a professional setting:\n\n'

    prompt_append = f'\n\nFrom the above text, extract any mention of {check_contains}, or say "none" if there aren\'t any:'
    prompt = prompt_prepend + text + prompt_append
    # TODO: check length of `text`` and truncate

    # Run rerun on high logprobs outputs from first run, need to catch b/c model likes to say none
    contains, reason = run_and_postprocess(prompt,
                                           return_extra=return_reason,
                                           run_rerun=True,
                                           uses_bool_none=True,
                                           )

    if return_reason:
        return contains, reason
    else:
        return contains


if __name__ == "__main__":
    texts = ['I like carrots and llamas.',
             'I dream of deforesting rainforests so that we have enough trees to capture all the carbon that we are all emitting.',
             '''CX: I want a taco with carnitas please.
             AI: Great, would you want a drink with that?
             CX: Actually, yeah, a large coke would be great. 
             ''',
             ]

    checks_contains = ['an orange item',
                       'climate change',
                       'upsell attempt by the AI'
                       ]

    for text, check_contains in zip(texts, checks_contains):
        is_contained, reason = fuzzy_contains(
            text, check_contains, return_reason=True)

        print(f'Text: {text}')
        print(f'Is {check_contains} in the text? {"Yes" if is_contained else "No"}')
        print(f'If yes, the reason is {reason}')
        print('=============================')



import requests
import json
from powerml_app.utils.run_ai import powerml_completions
from typing import TypedDict


class ConfigType(TypedDict):
    key: str


class PowerML:
    def __init__(self, config: ConfigType):
        self.config = config
        self.key = self.config["key"]
        print(self.key)

    def predict(self,
                prompt: str = "Say this is a test",
                stop: str = "",
                model: str = "llama",
                max_tokens: int = 128,
                temperature: int = 0,
                ) -> str:
        params = {
            "prompt": prompt,
            "model": model,
            "max_tokens": max_tokens,
            "stop": stop,
            "temperature": temperature,
        }
        # if the model is one of our models, then hit our api
        try:
            resp = powerml_completions(params, self.key)
        except:
            raise Exception("API failing")
        if 'error' in resp:
            raise Exception(str(resp))
        text = resp['completion']
        return text

    def fit(self,
            data: list[str],
            model: str = "llama"):
        # Upload filtered data to train api
        response = requests.post(
            url="https://api.staging.powerml.co/v1/train",
            json={
                "dataset": self.__make_dataset_string(data),
                "model": model
            })
        return response.json()

    def __make_dataset_string(self, training_data):
        dataset = "\n".join(
            [json.dumps({"prompt": item.replace("\n", "\\n")}) for item in training_data])
        print("Generated dataset: ", dataset)
        return dataset



import sys
import os


if __name__ == "__main__":
    path = os.path.join(os.path.dirname(os.path.dirname(
        os.path.realpath(__file__))), "src")
    sys.path.append(path)
    from powerml_app.utils.run_ai import run_ai
    response = run_ai()
    print(response)

    from powerml_app.PowerML import PowerML
    config = {"key": ""}
    powerml = PowerML(config)
    testPrompt = "hello there"
    response = powerml.predict(prompt=testPrompt)
    print(response)

    data = ["item1\n new line", "item2", "item3"]
    val = powerml.fit(data, "llama")
    print(val['model'])



'''Utility functions'''

from powerml_app.utils.run_ai import run_ai
from powerml_app.utils.constants import MAX_SHORT_TOKENS, N_LOGPROBS, LOGPROB_THRESHOLD


def get_top_tokens_by_logprobs(logprobs, uses_bool_none=True):
    '''Gets the important tokens, based on logprobs. If it's for a bool check, filters out None's.'''
    filtered_tokens = []
    for t in logprobs['top_logprobs']:
        if '<|endoftext|>' in t and t['<|endoftext|>'] > LOGPROB_THRESHOLD:
            break

        top_tokens = t.keys()
        for top_t in top_tokens:
            if t[top_t] < LOGPROB_THRESHOLD:
                continue

            if '\n' in top_t:
                continue
            if uses_bool_none and ('none' in top_t.lower() or top_t.lower() == 'n'):
                continue

            filtered_tokens.append((top_t, t[top_t]))
    return sorted(filtered_tokens, key=lambda ft: ft[1], reverse=True)


def postprocess_bool_none(output):
    '''Postprocesses boolean outputs with "none" for boolean output, and explanation/reason as extra output.'''
    bool_check = True
    output = output.strip()
    if output.lower() == 'none':
        bool_check = False
        output = output.lower().replace('none', '')
    elif 'none' in output.lower():
        print(
            f'Not sure if none is it b/c other words are outputted. TODO: run another LLM on it to double check. Output is {output}')
        bool_check = False
        output = output.lower().replace('none', '')

    extra_output = postprocess_helper_extra_output(output)
    return bool_check, extra_output


def postprocess_bool_yes_no(output):
    '''Postprocesses boolean outputs with yes/no.'''
    bool_check = True
    output = output.strip()
    if 'no' in output.lower():
        bool_check = False
    extra_output = postprocess_helper_extra_output(output)
    return bool_check, extra_output


def postprocess_bool_bullet_explanations(output):
    '''Postprocesses boolean outputs with yes/no and bullet explanations'''
    bool_check = True
    output = output.strip().split('\n')
    extra_output = []
    for i, o in enumerate(output):
        if i == 0:
            if 'no' in o.lower():
                bool_check = False
        elif '-' in o:
            o = o.split('-')[-1]
            extra_output.append(o.strip())
        else:
            print(f'Skipping processing this output {o}')
    return bool_check, extra_output


def postprocess_helper_extra_output(output):
    extra_output = []
    for o in output.split('\n'):
        if o:
            o_comma_split = [oo for oo in o.split(',') if oo]
            extra_output.extend(o_comma_split)
    if extra_output == ['']:
        extra_output = []
    return extra_output


def run_and_postprocess(prompt,
                        return_extra=False,
                        return_top_tokens=False,
                        uses_bool_none=False,
                        uses_bool_bullet=False,
                        run_rerun=False,
                        rerun_prompt_append=None,
                        max_tokens=MAX_SHORT_TOKENS,
                        ):
    '''
    Run LLM and postprocesses output.
    run_rerun: In a setting where we rerun with previous iteration's high logprobs.
    rerun_prompt_append: If run_rerun is True, use this extra bit in prompt
    '''
    if run_rerun and rerun_prompt_append:
        prompt += f' {rerun_prompt_append.strip()}'
        output = run_ai(prompt=prompt,
                        max_tokens=max_tokens,
                        return_logprobs=False,
                        )
        rerun_output = rerun_prompt_append + output
        print(f'Rerun output is: {rerun_output}')
    else:
        output, logprobs = run_ai(prompt=prompt,
                                  max_tokens=max_tokens,
                                  n_logprobs=N_LOGPROBS,
                                  return_logprobs=True,
                                  )

    if uses_bool_none:
        bool_check, extra_output = postprocess_bool_none(output)
    elif uses_bool_bullet:
        bool_check, extra_output = postprocess_bool_bullet_explanations(output)
    else:
        bool_check, extra_output = postprocess_bool_yes_no(output)

    is_check_high_logprobs = run_rerun and not bool_check and rerun_prompt_append is None

    if not logprobs is None:
        if return_top_tokens or is_check_high_logprobs:
            high_tokens_logprobs = get_top_tokens_by_logprobs(
                logprobs, uses_bool_none)

            # Check for high logprob tokens that could've been missed, in original run (not rerun)
            if is_check_high_logprobs:
                print(
                    f'Tokens with important logprobs that were not sampled: {high_tokens_logprobs}')

                if high_tokens_logprobs:
                    # TODO: Check all of them, not just the first one (best one, b/c sorted)
                    rerun_prompt_append = high_tokens_logprobs[0][0]
                    bool_logprob_check, extra_output_logprob_check = run_and_postprocess(
                        prompt,
                        return_extra=True,
                        return_top_tokens=False,
                        run_rerun=True,
                        rerun_prompt_append=rerun_prompt_append
                    )

                    # Override contains if applicable
                    bool_check = max(bool_check, bool_logprob_check)
                    if bool_check:
                        rerun_output = rerun_prompt_append
                        if extra_output_logprob_check:
                            rerun_output += extra_output_logprob_check[0]
                        extra_output.append(rerun_output)

    returns = [bool_check]
    if return_extra:
        returns.append(extra_output)

    if return_top_tokens and not logprobs is None:
        returns.append(high_tokens_logprobs)

    if len(returns) > 1:
        return returns
    else:
        return returns[0]



import requests

import logging

logger = logging.getLogger(__name__)


def run_ai(prompt="Say this is a test",
           stop="",
           model="llama",
           max_tokens=128,
           return_logprobs=False,
           api="powerml",
           temperature=0,
           key="",
           ):

    params = {
        "prompt": prompt,
        "model": model,
        "max_tokens": max_tokens,
        "stop": stop,
        "temperature": temperature,
    }
    if api == "powerml":
        # if the model is one of our models, then hit our api
        try:
            resp = powerml_completions(params, key)
        except:
            raise Exception("API failing")
        if 'error' in resp:
            raise Exception(str(resp))
        text = resp['completion']
        if return_logprobs:
            return text, None
        return text
    else:
        # otherwise hit openai
        try:
            resp = openai_completions(params)
        except:
            raise Exception("Run 'powerml start' first")
        if 'error' in resp:
            raise Exception(str(resp))
        text = resp['choices'][0]['text']
        if return_logprobs:
            logprobs = resp['choices'][0]['logprobs']
            return text, logprobs
        return text


def powerml_completions(params, key):
    headers = {
        "Authorization": "Bearer " + key,
        "Content-Type": "application/json", }
    response = requests.post(
        url="https://api.staging.powerml.co/v1/completions",
        headers=headers,
        json=params)
    return response.json()


def openai_completions(params, key):
    headers = {
        "Authorization": "Bearer " + key,
        "Content-Type": "application/json", }
    response = requests.post(
        url="https://api.openai.com/v1/completions",
        headers=headers,
        json=params)
    return response.json()


if __name__ == "__main__":
    text = run_ai()
    print(text)



import functools
import csv
import os


def powerml_data(func):
    @functools.wraps(func)
    def wrapper():
        print("\nGenerating_data: " + func.__name__)
        result = func()
        print(result)
        return result
    wrapper.is_generator_function = True
    return wrapper


def powerml_data_local(func):
    @functools.wraps(func)
    def wrapper():
        print("\nGenerating data and saving to file: " + func.__name__)
        generated_data = func()
        print(generated_data)
        # save generated data to a csv file
        cwd = os.getcwd()
        data_cwd = cwd + "/data"
        with open(data_cwd + '/data.csv', 'w', newline='') as csvfile:
            fieldnames = ['model_input', 'model_output']
            writer = csv.DictWriter(csvfile, fieldnames=fieldnames)
            writer.writeheader()
            for data in generated_data:
                writer.writerow(data)
        return generated_data
    wrapper.is_generator_function = True
    return wrapper



import functools


def powerml_filter(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        print("\nRunning Filter: " + func.__name__)
        # Load Data to be filtered
        data_to_filter = kwargs['data']
        passing_data = []
        # pass data into filter function
        for datapoint in data_to_filter:
            result = func(datapoint["model_output"])
            if result:
                passing_data.append(datapoint)
        print("Cases: %d" %
              (len(data_to_filter)))
        print("Result: %d%%" %
              (100*len(passing_data)/len(data_to_filter)))
        return passing_data
    wrapper.is_filter_function = True
    return wrapper



from powerml_app.utils.run_ai import run_ai
import re


def more_sql_like_this(example, n=1):
    '''
    Uses LLM to get more queries like the example.
    '''
    prompt_append = f'Write %d more example(s) that use the same schema, but answer a different business question than the above. List them below.\n\nExample 1. ' % (
        n)
    prompt = f'Example: {example}\n\n{prompt_append}'
    text = run_ai(prompt,
                  max_tokens=256,
                  )
    results = re.split("Example \d+\.", text)
    results = [strip_end_of_query(r).strip() for r in results]
    return results


def strip_end_of_query(query):
    location = query.find(";")

    if location > 0:
        return query[:location+1]

    return query



from powerml_app.utils.utils import run_and_postprocess


def fuzzy_duplicate_filter(text, return_duplicates=False):
    '''
    Uses LLM to check whether text has nonsensical duplicates (model is repeating itself).

    text: Input text to check.
    '''
    prompt_append = f'\n\nFrom the above text, extract any repetitive phrases that makes no sense. If none, write "none".'
    prompt = text + prompt_append

    bool_check, extra = run_and_postprocess(prompt,
                                            return_extra=return_duplicates,
                                            uses_bool_none=True,
                                            )

    if return_duplicates:
        return bool_check, extra
    else:
        return bool_check


if __name__ == "__main__":
    texts = [
        'I like carrots and llamas.',
        '''CX: I'd like to order.
AI: Sorry we are out of that.
CX: Sorry we are really really out of that.
Order: 1x Vanilla Syrup''',
    ]

    for text in texts:
        bool_check, duplicates = fuzzy_duplicate_filter(
            text, return_duplicates=True)

        print(f'Text: {text}')
        print(
            f'Are there duplicates that make no sense? {"Yes" if bool_check else "No"}')
        print(f'If yes, the reason is {duplicates}')
        print('=============================')
